import calendar
from collections import defaultdict
import datetime
import re
import numpy as np

import pandas as pd

from .config import *


class OperatingUnit:
    def __init__(
        self,
        worksheet,
        operating_unit,
        year,
        name=None,
        budgeted_account=None,
    ) -> None:
        self.worksheet = worksheet
        self.operating_unit = operating_unit
        self.name = name
        self.year = year

        if budgeted_account is None:
            self.budgeted_account = operating_unit.startswith("R")
        else:
            self.budgeted_account = budgeted_account

        # Create a new dataframe with months of the year as column headers
        self.df_gen = pd.DataFrame(
            columns=("Type", *[calendar.month_abbr[i] for i in range(1, 13)])
        )

        self.row_colors = {}
        self.hidden_rows = []

        # Default options
        self.set_options()

        self.df = None

        if self.budgeted_account:
            self.previous_expenditures = self.analyzer.get_previous_expenditures(
                self.operating_unit, self.year - 1
            )

    def set_options(self, hide_student_wages=False):
        self.hide_student_wages = hide_student_wages

    def run(self, df):
        self.df = df
        if self.year:
            # Filter df on 'Fiscal Year' column
            self.df = self.df[self.df["Fiscal Year"] == self.year]

        if self.budgeted_account:
            self.verify_no_income()
        else:
            self.filter_out_budgeted_data()

        print(
            f"Running Operating Unit:{self.operating_unit} {('(' + str(self.year) + ')') if self.year else ''}",
        )

        # Add student wage data
        self.add_row("Student Wages", CODES_STUDENT_WAGES, color=COLOR_STUDENT_WAGES)
        self.create_student_wages()
        self.add_row("Student Benefits", CODES_STUDENT_BENEFITS, color=COLOR_STUDENT_BENEFITS)
        self.add_row("Student Tuition", CODES_STUDENT_TUITION, color=COLOR_STUDENT_TUITION)
        self.add_row("Faculty Spr/Sum", CODES_FACULTY_SPRING_SUMMER, color=COLOR_FACULTY)
        self.add_row("Travel", CODES_TRAVEL, color=COLOR_TRAVEL)
        self.add_row("Supplies/Misc", CODES_SUPPLIES, color=COLOR_SUPPLIES)
        self.add_row("Capital Equipment", CODES_CAPITAL, color=COLOR_CAPITAL)
        self.add_row("Overhead", CODES_OVERHEAD, color=COLOR_OVERHEAD)
        self.row_total_expenses = self.add_row(
            "Total Expeneses", all_expense_codes(), color=COLOR_TOTAL_EXPENSES
        )

        self.add_empty_row()

        if not self.budgeted_account:
            self.add_row("Interest", CODES_INTEREST, actuals_coeff=-1)

        self.add_row(
            "New Budgets or Carryover" if self.budgeted_account else "Transfer Income / Carryover",
            CODES_INCOME,
            budgets_coeff=1,
            actuals_coeff=-1,
        )

        self.row_total_income = self.add_row(
            "Total Income",
            CODES_INTEREST + CODES_INCOME,
            color=COLOR_TOTAL_INCOME,
            budgets_coeff=1,
            actuals_coeff=-1,
        )
        self.add_empty_row()

        # Add a sum column that sums the values in each row, starting with the 2nd column,
        # skipping rows where "Type" is empty

        # Calculate the sum for rows where 'Type' column is not empty
        columns_to_sum = self.df_gen.columns[1:]  # Get all columns starting from the second column
        self.df_gen.loc[self.df_gen["Type"] != "", "Sum"] = self.df_gen.loc[
            self.df_gen["Type"] != "", columns_to_sum
        ].sum(axis=1)

        # Save total expenses, which is found in the self.row_total_expenses, "Sum" column
        self.total_expenses = self.df_gen.iloc[self.row_total_expenses]["Sum"]

        self.add_balance_row()
        self.add_empty_row()

        title_name = self.operating_unit
        if self.name:
            title_name = f"{self.name} ({title_name})"
        if self.year:
            title_name = f"{title_name} - {self.year}"
        self.worksheet.write_to_sheet(self.df_gen, title_name, self.hidden_rows, self.row_colors)

    def add_empty_row(self):
        # Add an empty row of None values
        empty_data = {
            "Type": [""],
            **{calendar.month_abbr[month]: [np.nan] for month in range(1, 13)},
        }

        self.df_gen = pd.concat((self.df_gen, pd.DataFrame(empty_data)), ignore_index=True)

    def add_row(
        self,
        row_name,
        account_codes,
        budgets_coeff=0,
        actuals_coeff=1,
        color=None,
    ):
        """Sum the values in the 'Amount' column for each month in the given account codes.
        If actual=True, then only actual values will be summed.
        Otherwise budgeted (account income) values will be summed."""

        new_row_data = {
            "Type": [row_name],
            **{
                calendar.month_abbr[month]: [
                    budgets_coeff
                    * (
                        self.df.loc[
                            (self.df["Accounting Period"] == month)
                            & (self.df["JRNL Line Ledger"] == "BUDGETS")
                        ][COL_NAME_AMOUNT].sum()
                    )
                    + actuals_coeff
                    * (
                        self.df.loc[
                            (self.df["Accounting Period"] == month)
                            & self.df["Account"].isin(account_codes)
                            & (self.df["JRNL Line Ledger"] == "ACTUALS")
                        ][COL_NAME_AMOUNT].sum()
                    )
                ]
                for month in range(1, 13)
            },
        }

        # Add this dictionary to the dataframe
        self.df_gen = pd.concat((self.df_gen, pd.DataFrame(new_row_data)), ignore_index=True)

        if color:
            self.row_colors[len(self.df_gen)] = color

        return self.df_gen.tail(1).index[0]

    def add_balance_row(self):
        if self.budgeted_account:
            self.df_gen.loc[len(self.df_gen)] = {
                "Type": "Previous Expenditures",
                calendar.month_abbr[1]: self.previous_expenditures,
            }

        self.df_gen.loc[len(self.df_gen)] = {"Type": "Balance (end of month)"}

        # # Iterate through columns and calcuate the balance, which is the total income minus the total expenses
        for month in range(1, 13):
            income = self.df_gen.iloc[self.row_total_income][month]
            expenses = self.df_gen.iloc[self.row_total_expenses][month]

            if month > 1:
                prev_balance = self.df_gen.iloc[len(self.df_gen) - 1, month - 1]
            elif self.budgeted_account:
                prev_balance = -self.df_gen.iloc[len(self.df_gen) - 2, month]
            else:
                prev_balance = 0

            balance = income - expenses + prev_balance
            self.df_gen.iloc[len(self.df_gen) - 1, month] = balance
            if month == 12:
                self.balance = balance

        self.row_colors[len(self.df_gen)] = "yellow"

    def create_student_wages(self):
        # Filter to only student wages ('Account' column in STUDENT_WAGES)
        df_students = self.df[
            (self.df["Account"].isin(CODES_STUDENT_WAGES))
            & (self.df["JRNL Line Ledger"] == "ACTUALS")
        ]

        # Get unique list of value in 'JRNL Line Descr' column
        students_raw = df_students["JRNL Line Descr"].unique()
        self.students = defaultdict(list)
        for student in students_raw:
            match = re.search("([A-Za-z]+)[, ]+([A-Za-z]+)", student)
            if not match:
                raise Exception(f"Could not parse student name: {student}")
            student_key = f"{match.group(2).title()} {match.group(1).title()}"
            self.students[student_key].append(student)

        # Create a new dictionary with the student wages by month
        student_data = {
            "Type": [f"  {s}" for s in sorted(self.students.keys())],
            **{
                calendar.month_abbr[month]: [
                    df_students.loc[
                        (df_students["Accounting Period"] == month)
                        & (df_students["JRNL Line Descr"].isin(self.students[student]))
                    ]["JRNL Monetary Amount -no scrn aggregation"].sum()
                    for student in sorted(self.students.keys())
                ]
                for month in range(1, 13)
            },
        }
        self.df_gen = pd.concat((self.df_gen, pd.DataFrame(student_data)), ignore_index=True)

        # Add new row index to hidden rows
        if self.hide_student_wages:
            self.hidden_rows.extend(
                range(len(self.df_gen) - len(student_data["Type"]) + 1, len(self.df_gen) + 1)
            )

    def verify_no_income(self):
        # Ensure df contains no CODES_INCOME
        if self.df[self.df["Account"].isin(CODES_INCOME + CODES_INTEREST)].shape[0] > 0:
            raise Exception(
                f"Budgeted account {self.operating_unit} contains income. "
                "Budgeted accounts should only contain expenses."
            )

    def filter_out_budgeted_data(self):
        # Filter out budgeted data.  Only include ACTUALS or BUDGET with code in CODES_NON_BUDGETED_BUDGET_CODES
        self.df = self.df[
            (
                (self.df["JRNL Line Ledger"] == "ACTUALS")
                | (
                    (self.df["JRNL Line Ledger"] == "BUDGETS")
                    & (self.df["Account"].isin(CODES_NON_BUDGETED_BUDGET_CODES))
                )
            )
        ]

    @property
    def analyzer(self):
        return self.worksheet.workbook.analyzer



CODES_STUDENT_WAGES = (
    5510, # Contracts: Student teaching assist salary
    5599, # Contracts: Student Asst Salary (Reclass)
    5572, # Contracts: Stdnt Research Assist Sal-FICA
    5570, # Contracts: Student Research Assist Salary
    5610, # Time Cards: Student Wages-Teaching Assist
    5672, # Time Cards: Stdnt Wages Rsrch Assist-FICA
    5600, # Time Cards: Student Wages
    5602, # Time Cards: Student Wages-FICA
    5720, # Non-Student Wages: Category 1 no Benefit
    5750, # Non-Student Wages: Category 2 Part Benefit
    8905, # Sponsored Research: Research Participant Stipends
    5812, # ?
)

# These student benefits aren't attached to an individual student
# name, so they must be separated out
CODES_STUDENT_BENEFITS = (
    5670, # Time Cards: Student Wages-Research Assist
    5699, # Time Cards: Student Wages (Reclass)
    5950, # Student Contract Benefits
    5960, # Student Time Card Benefits
    5970, # Non Student Benefits
    5999, # Benefits (reclass)
)

CODES_STUDENT_TUITION = (
    6300, # Scholarships
    6303, # 
    6304, # 
    6309, # 
    6319, # Student Aid (reclass) 
    6390, # Other Student Aid
)

CODES_FACULTY_SPRING_SUMMER = (
    5920, # Faculty Supplemental Benefits
    5260, # Faculty Sal-Spr/Sum-Benefits
    5220, # Faculty Sal-Spr/Sum No Ben
    5240, # Faculty Sal-Supp (Overload)
)

CODES_TRAVEL = (
    6190, # Employee Development/Training
    6400, # Insurance Expense
    6494, # Other Medical
    7000, # Domestic-Business Travel
    7010, # Domestic-Recruiting Travel
    7020, # Domestic-Prof Developmt Travel
    7030, # Domestic-Student Travel
    7050, # Foreign-Business Travel
    7060, # Foreign-Recruiting Travel
    7070, # Foreign-Prof Developmt Travel
    7080, # Foreign-Student Travel
)

CODES_SUPPLIES = (
    2200, #
    5980, # Other Payroll Benefits
    6000, # Software Acquisitions/Support
    6005, # Software / Media Licenses
    6100, # Supplies
    6109, # Supplies (reclass)
    6110, # Supplies-BYU Store
    6120, # Printing and Copying
    6125, # Research Publications Costs
    6130, # Postage and Mailing
    6140, # Telecommunications
    6160, # Advertising and Promotion
    6180, # Hosting, Food, Entertainment
    6185, # Guest Tickets
    6192, # Employee Memberships, Dues
    6200, # Contract Service-Off Campus
    6210, # Contract Services-Campus
    6220, # Rental Expense - Intra Campus
    6250, # Non-capital equipment
    6255, # Non-Cap Access Equipment
    6270, # Equipment Maintenance
    6403, # Other Academic Payments
    6405, # Prizes
    6600, # Access Equipment
    6610, # Access Lab Equipment
    6480, # Legal Fees
    6490, # Other Expense
    6499, # Other Expense (reclass)
    8940, # Conference Hosting (R-project)
)

CODES_CAPITAL = (
    1625, #
    1725, # 
    8930, # Research Capital Equipment
    9126, # AM Retirement-Assets
    9127, # AM Retirement-Accum Depr
    9136, # AM Transfer-Assets
    9137, # AM Tansfer-Accum Depr
)

CODES_OVERHEAD = (
    8990, # Facilities & Admin Recovery
    8991, # Waived Facilities & Admin Cost
    8993, # Subcontract Facilities & Admin
)

CODES_INTEREST = (
    4711, # Interest Alloc-Claim on Cash
    4719, # Non-Invest-Intrst Inc(reclass)
    # These probably need to go to another location. They are put here temporarily
    4755, # Distributed Royalty Income
    4891, # Endowment Payout (UFS only)
)

# Income only applies to non-budgeted accounts
CODES_INCOME = (
    8920, # Subawards w/ F&A
    8925, # Subawards w/out F&A
    9250, # Transfer Into Net Assets
    9260, # Transfer Out of Net Assets
    3500, # 
)

# Non-budgeted accounts, don't normally use BUDGET lines and they
# can be removed, except for these codes
CODES_NON_BUDGETED_BUDGET_CODES = (10,)

# These codes aren't used by anything and can be removed from the data
CODES_IGNORED = (
    1010, #
    1235, #
    2000, #
    2210, #
    4200, # Gift Revenue-Cash
    4220, # Grants & Contract Revenue
)

COL_NAME_AMOUNT = "JRNL Monetary Amount -no scrn aggregation"

COLOR_STUDENT_WAGES = "rosybrown"
COLOR_STUDENT_BENEFITS = "moccasin"
COLOR_STUDENT_TUITION = "moccasin"
COLOR_FACULTY = "lightskyblue"
COLOR_TRAVEL = "plum"
COLOR_SUPPLIES = "khaki"
COLOR_CAPITAL = "khaki"
COLOR_OVERHEAD = "silver"
COLOR_TOTAL_EXPENSES = "tomato"
COLOR_TOTAL_INCOME = "limegreen"


def all_expense_codes():
    return (
        CODES_STUDENT_WAGES
        + CODES_STUDENT_BENEFITS
        + CODES_STUDENT_TUITION
        + CODES_FACULTY_SPRING_SUMMER
        + CODES_TRAVEL
        + CODES_SUPPLIES
        + CODES_OVERHEAD
        + CODES_CAPITAL
    )


def all_handled_codes():
    return all_expense_codes() + CODES_INTEREST + CODES_IGNORED + CODES_INCOME



import pathlib
import pandas as pd

from .config import CODES_IGNORED, COL_NAME_AMOUNT, all_expense_codes, all_handled_codes
from .utils import filter_df_on_operating_units
from .workbook import Workbook


class CsvDataAnalyzer:
    def __init__(self, csv_path, split_notebooks=False) -> None:
        self.csv_path = csv_path
        self.split_notebooks = split_notebooks

        self.df = pd.read_csv(self.csv_path)

        # Convert Operating Unit to string
        self.df["Operating Unit"] = self.df["Operating Unit"].apply(str)

        # Filter out ignored codes (account in CODES_IGNORED)
        self.df = self.df[~self.df["Account"].isin(CODES_IGNORED)]

        self.operating_units = sorted([str(ou) for ou in self.df["Operating Unit"].unique()])
        print(f"Operating Units Found in CSV: {', '.join(self.operating_units)}")

        self.workbooks_initialized = False
        self.workbooks = []

        self.check_unhandled_codes()

    def add_workbook(self, path):
        self.workbooks_initialized = True
        workbook = Workbook(path)
        workbook.analyzer = self
        self.workbooks.append(workbook)
        return workbook

    def _setup_default_workbooks(self):
        if self.split_notebooks:
            # Create a new workbook for each operating unit
            for operating_unit in self.operating_units:
                self.add_workbook(pathlib.Path.cwd() / (operating_unit + ".xlsx")).add_worksheet(
                    operating_unit
                ).add_operating_unit(operating_unit)
        else:
            # Create a single workbook with all operating units
            wb = self.add_workbook(pathlib.Path.cwd() / "all.xlsx")
            for operating_unit in self.operating_units:
                wb.add_worksheet(operating_unit).add_operating_unit(operating_unit)

    def run(self):
        if not self.workbooks_initialized:
            self._setup_default_workbooks()

        for workbook in self.workbooks:
            workbook.run(filter_df_on_operating_units(self.df, workbook.get_operating_units()))

    def check_unhandled_codes(self):
        # Get all account codes that were not handled
        unhandled_codes = sorted(
            self.df[
                (~self.df["Account"].isin(all_handled_codes()))
                & (self.df["JRNL Line Ledger"] == "ACTUALS")
            ]["Account"].unique()
        )

        # TODO: Add to spreadsheet instead of raising exception

        if unhandled_codes:
            raise Exception(
                f"Unhandled account codes: {','.join([str(c) for c in unhandled_codes])}.  See https://finserve.byu.edu/account-revenue-expense"
            )

    def get_all_operating_units(self):
        return [
            ou
            for workbook in self.workbooks
            for worksheet in workbook.worksheets
            for ou in worksheet.operating_units
        ]

    def find(self, operating_unit, year=None):
        operating_uints = [
            ou for ou in self.get_all_operating_units() if ou.operating_unit == str(operating_unit)
        ]
        if year:
            operating_uints = [ou for ou in operating_uints if ou.year == year]
        return operating_uints

    def get_previous_expenditures(self, operating_unit, year):
        """Return the sum of expenditures for this year and all previous years"""
        return self.df[
            (self.df["Operating Unit"] == operating_unit)
            & (self.df["Fiscal Year"] <= year)
            & (self.df["JRNL Line Ledger"] == "ACTUALS")
            & (self.df["Account"].isin(all_expense_codes()))
        ][COL_NAME_AMOUNT].sum()
